/* 
Version 0.01
The license of this file is POVray-ware, that is use of this file and/or any part of it is free, as long it is used directly in POVray or one of its unofficial rebuilds.
Contents/ syntax: 
  -Matricial sum 
           syntax:  summat(array1,array2,array3), 
           effect:    array3=array1+array2 
 
  -Scalar product 
           syntax:  scalprod(array1,array2,scalar)
           effect:    array2=scalar*array1
 
  -Transpose 
           syntax:  transp(array1,array2)                
           effect:    array2=array1' ('=transposed)
 
  -Matricial product 
           syntax:  matprod(array1, array2, array3)              
           effect:    array3=array1*array2
 
  -Triple matricial product (useful for matrix inversion algorithm, henceforth UFMIL)
           syntax:  triplematprod(array1,array2,array3,array4)             
           effect:    array4=((array1*array2)*array3)
 
  -Quadratic form 
           syntax:  quadform(vec1,array1,total)      
           effect:    total=(vec1)'*array1*vec1
 
  -Row extract 
           syntax:  rowextr(array1,array2,rown)     
           effect:    array2 (row vector)=row number "rown" of array1, rown=1 means first row, etc.
 
  -Column extract
           syntax:  colextr(array1,array2,coln)
           effect:    array2 (column vector)=column number "coln" of array1, coln=1 means first column, etc.
 
  -Identity matrix 
           syntax:  identmat(matrixname, rank)
           effect:    generates I(rank), a matrix filled with 1's on the main diagonal and 0's elsewhere
 
  -Matrix fill 
           syntax:  matrixfill(matrixname,m,n,value)      
           effect:    fills a m*n matrix ("matrixname") with constant "value"
 
  -Random fill:
           syntax:  randomfill(matrixname,m,n,seedvalue)
           effect:    fills a m*n matrix with random values generated with seed=seedvalue
 
  -Matrix subextract:
           syntax:  subextract(array1,array2,a,b,c,d)      
           effect:    array2 is the matrix formed from extracting the values starting at row a, column b, to row c column d. a may be greater than c, and b of d.
 
  -Find maximum value (beta, may change):
           syntax:  matrixmax(array1)
           effect:    stores the highest value of the matrix in variable "maxi", also prints in message window the coordinates of the max
 
  -Find minimum value (beta, may change):
           syntax:  matrixmin(array1)
           effect:    stores the lowest value of the matrix in variable "mini", also prints in message window the coordinates of the min   
 
  -ZZLY Special Bordered matrix (UFMIL)
           syntax:  zzlyborder(array1,array2)         
           effect:    array2 is array1 plus an extra column and a row filled with 0's, except the intersection of the extra row and column has a value of 1
 
  -Bordered Matrix:        
           syntax:  bordermatrix(array1,array2,array3,array4,array5)
           effect:            array1 | array2
                      array5= -------+-------
                              array3 | array4  ; #rows of array1 must equal #rows of array2, etc.
 
  -Matricial equality (UFMIL):
           syntax:  matrixcond(array1,array2)
           effect:    tests array1=array2 equality, the scalar "matrixcond1" will equal 1 if they are equal, 0 otherwise
 
  -Row append: to-do
 
  -Moore-Penrose generalised inverse:
           syntax:  MPGinverse(array1,array2)
           effect:    array2=(array1)^(-1); can invert singular and even rectangular matrices    
                  
*/  
// matrix sum 
#macro summat(array1, array2,array3)
   #local dime=0;
   #local a1=dimension_size(array1,1);
   #if (a1=dimension_size(array2,1))    
       #local a2=dimension_size(array1,2);
       #if (a2=dimension_size(array2,2))
            #local i=0;
            #declare array3= array[a1][a2];
            #while (i<=a1-1)
                  #local j=0;
                  #while (j<=a2-1)
                       #declare array3[i][j]=array1[i][j]+array2[i][j]; 
                       #local j=j+1;
                  #end              
                  #local i=i+1;
            #end
        #else #local dime=1;
        #end
    #else #local dime=1;
    #end
    #if (dime=1) #debug "********Dimension mismatch (Sum)********\n" 
    #end
#end

// scalar product 
#macro scalprod(array1,array2,scalar)                  
     #local a1=dimension_size(array1,1);     
     #local a2=dimension_size(array1,2);
     #local i=0;
     #declare array2= array[a1][a2];
     #while (i<=a1-1)
           #local j=0;
           #while (j<=a2-1)
                 #declare array2[i][j]=array1[i][j]*scalar;
                 #local j=j+1;
           #end              
           #local i=i+1;
     #end
#end

// transpose 
#macro transp(array1,array2)
     #local a1=dimension_size(array1,1);     
     #local a2=dimension_size(array1,2);
     #local i=0;
     #declare array2= array[a2][a1];
     #while (i<=a2-1)
           #local j=0;
           #while (j<=a1-1)
                  #declare array2[i][j]=array1[j][i];
                  #local j=j+1;
           #end              
           #local i=i+1;
     #end
#end   

// matricial product
#macro matprod(array1, array2, array3)
   #local matm=dimension_size(array1,2);
   #if (matm=dimension_size(array2,1))
       #local mato=dimension_size(array1,1);
       #local matn=dimension_size(array2,2);
            #local i=0;
            #declare array3= array[mato][matn];
            #while (i<=mato-1)
                  #local j=0;
                  #while (j<=matn-1)
                        #local summ=0;
                        #local k=0;
                              #while (k<=matm-1)
                                    #local summ=summ+array1[i][k]*array2[k][j];
                                    #local k=k+1;
                              #end      
                       #declare array3[i][j]=summ;
                       #local j=j+1;
                  #end              
                  #local i=i+1;
            #end

    #else 
    #debug "********Dimension mismatch (Product)********\n"
    #end 
#end

// triple product 
#macro triplematprod(array1,array2,array3,array4)
 #local matm=dimension_size(array1,2);
 #if (matm=dimension_size(array2,1))
    matprod(array1,array2,temp1)
    #local matn=dimension_size(temp1,2);
    #if (matn=dimension_size(array3,1))
       matprod(temp1,array3,array4)
    #else
         #debug "********Dimension mismatch (Triple Product)********\n"
    #end        
 #else 
       #debug "********Dimension mismatch (Triple Product)********\n"
 #end                                                       
#end
                                        
// quadratic forms vec1 is a column vector, array1 the square matrix, and total the name of the (scalar) result                                       
#macro quadform(vec1,array1,total) 
 #local vecm=dimension_size(vec1,2);
 #if (vecm=1)
     #local vecn=dimension_size(vec1,1); 
     #if (vecn=dimension_size(array1,1)) 
         #local i=0;
         #declare total=0;
         #while (i<=vecn-1)
             #local j=0;
             #while (j<=vecn-1)
                #declare total=total+array1[i][j]*vec1[i][0]*vec1[j][0];
                #local j=j+1;
             #end
         #local i=i+1;
         #end
      #else #debug "********Dimension mismatch (Quadratic Form)********\n"              
      #end
  #else #debug "********First argument needs to be a vector**********\n"    
  #end     
#end

// row extract 
#macro rowextr(array1,array2,rown) // rown is an integer: 1=first row, etc.
#local i=0;                          
#declare array2=array[1][dimension_size(array1,2)];
 #while (i<= dimension_size(array1,2)-1)
      #declare array2[0][i]=array1[rown-1][i];
      #local i=i+1;
 #end     
#end

/* column extract, array2 is the name of the column to be extracted from array1,
  coln is the ordinal of the column to be extracted: 1=first column, etc.*/
#macro colextr(array1,array2,coln)
#local i=0;                          
#declare array2=array[dimension_size(array1,1)][1];
 #while (i<= dimension_size(array1,1)-1)
      #declare array2[i][0]=array1[i][coln-1];
      #local i=i+1;
 #end     
#end


// Identity matrix
#macro identmat(matrixname, rank)
#declare matrixname=array[rank][rank];
#local i=0;
  #while (i<= rank-1) 
      #local j=0;
      #while (j<=rank-1)
            #if (i=j)
               #declare matrixname[i][j]=1;
            #else 
               #declare matrixname[i][j]=0;
            #end
      #local j=j+1;
      #end         
  #local i=i+1;      
  #end
#end

// Fill matrix m rows, n columns with "value" 
#macro matrixfill(matrixname,m,n,value)
#declare matrixname=array[m][n];
#local i=0;
  #while (i<= m-1) 
      #local j=0;
      #while (j<=n-1)
            #declare matrixname[i][j]=value;
      #local j=j+1;
      #end         
  #local i=i+1;      
  #end
#end
     
// Random fill
#macro randomfill(matrixname,m,n,seedvalue)
#declare matrixname=array[m][n];
#declare rr5=seed(seedvalue);
#local i=0;
  #while (i<= m-1) 
      #local j=0;
      #while (j<=n-1)
            #declare matrixname[i][j]=rand(rr5);
      #local j=j+1;
      #end         
  #local i=i+1;      
  #end
#end
     
/* matrix subextract,array1 is starting matrix, array2 the name of the extracted matrix, a,b row and column start ordinals and c,d row and column end ordinals.
If a > c they are swapped around, likewise for b and d.*/
#macro subextract(array1,array2,a,b,c,d) 
#if (a>c) 
   #local e=a;
   #local a=c;
   #local c=e;
#end
#if (b>d) 
   #local e=b;
   #local b=d;
   #local d=e;
#end   
#declare array2=array[c-a+1][d-b+1];
#local i=0;
  #while (i<= c-a) 
      #local j=0;
      #while (j<=d-b)
            #declare array2[i][j]=array1[a+i-1][b+j-1];
      #local j=j+1;
      #end         
  #local i=i+1;      
  #end
#end

// max value 
#macro matrixmax(array1)
#declare maxi=array1[0][0];  
#local m=dimension_size(array1,1);
#local n=dimension_size(array1,2);
#local i=0;
  #while (i<= m-1) 
      #local j=0;
      #while (j<=n-1)
            #if (array1[i][j]>maxi)
               #declare	maxi=array1[i][j];
               #local maxcoord1=i+1;
               #local maxcoord2=j+1;
            #end   
      #local j=j+1;
      #end         
  #local i=i+1;      
  #end 
  #debug concat("*********Matrix maximum: "str(maxi,1,0)" located at row "str(maxcoord1,-1,0)" and column "str(maxcoord2,-1,0)"*********\n" )
#end

// min value  
#macro matrixmin(array1)
#declare mini=array1[0][0];  
#local m=dimension_size(array1,1);
#local n=dimension_size(array1,2);
#local i=0;
  #while (i<= m-1) 
      #local j=0;
      #while (j<=n-1)
            #if (array1[i][j]<mini)
               #declare	mini=array1[i][j];
               #local mincoord1=i+1;
               #local mincoord2=j+1;
            #end   
      #local j=j+1;
      #end         
  #local i=i+1;      
  #end 
  #debug concat("*********Matrix minimum: "str(mini,1,0)" located at row "str(mincoord1,-1,0)" and column "str(mincoord2,-1,0)"*********\n" )
#end

// ZZLY-algorithm bordered matrix array1 is the matrix to be bordered, array2 the result 
#macro zzlyborder(array1,array2)
#local m=dimension_size(array1,1);
#local n=dimension_size(array1,2);
#declare array2=array[m+1][n+1];
#local i=0;
  #while (i<=m) 
      #local j=0;
      #while (j<=n)
            #if (i<m & j<n)
               #declare array2[i][j]=array1[i][j];
            #else 
               #declare array2[i][j]=0;
            #end   
      #local j=j+1;
      #end         
  #local i=i+1;      
  #end
#declare array2[m][n]=1;
#end    

// bordered matrix         array1 | array2
//                 array5= -------+-------
//                         array3 | array4
#macro bordermatrix(array1,array2,array3,array4,array5)
#local ar1m=dimension_size(array1,1);
#local ar1n=dimension_size(array1,2);
#local ar2m=dimension_size(array2,1);
#local ar2n=dimension_size(array2,2);
#local ar3m=dimension_size(array3,1);
#local ar3n=dimension_size(array3,2);
#local ar4m=dimension_size(array4,1);
#local ar4n=dimension_size(array4,2);
 #if (ar1m=ar2m & ar1n=ar3n & ar4m=ar3m & ar4n=ar2n)
    #declare array5=array[ar1m+ar3m][ar1n+ar2n];
    #local i=0;
    #while (i<ar1m+ar3m)
          #local j=0;
          #if (i<ar1m)
             #while (j<ar1n+ar2n) 
                   #if (j<ar1n)
                      #declare array5[i][j]=array1[i][j];
                   #else
                      #declare array5[i][j]=array2[i][j-ar1n];
                   #end   
             #local j=j+1;
             #end
          #else
             #while (j<ar1n+ar2n)
                    #if (j<ar1n)
                      #declare array5[i][j]=array3[i-ar1m][j];
                   #else
                      #declare array5[i][j]=array4[i-ar1m][j-ar1n];
                   #end 
             #local j=j+1;
             #end      
          #end 
    #local i=i+1;              
    #end      
  #else
  #debug "*********Dimension mismatch (Border matrix)*********\n" 
  #end
#end  

// matricial condition  array1<>array2
#macro matrixcond(array1,array2)
#if (dimension_size(array1,1)=dimension_size(array2,1) & dimension_size(array1,2)=dimension_size(array2,2))
   #local i=0;    
   #declare matrixcond1=1;
   #local k=0;
   #while (i<dimension_size(array1,1))
         #local j=0;
         #while (j<dimension_size(array1,2))
               #if (array1[i][j]=array2[i][j])
                  #local k=1;
               #else #local k=0;   
               #end
         #declare matrixcond1=matrixcond1*k;
         #local j=j+1;
         #end
   #local i=i+1;
   #end
   #if (matrixcond1=1) #debug "***********Matrices coincide***********\n" 
     #else #debug "***********Matrices differ***********\n" 
   #end             
 #else #debug "***********Dimension mismatch (Matrix Condition)***********\n"
 #end
#end

// Matrix (and column) append, array3=array1|array2
#macro colappend(array1,array2,array3)
#declare b=dimension_size(array1,1);
#declare a1=dimension_size(array1,2);
#declare a2=dimension_size(array2,2);
#if (b=dimension_size(array2,1))
   #local i=0;                          
   #declare array3=array[b][a1+a2];
   #while (i<= b-1)
         #local j=0;
         #while (j<=a1+a2-1) 
               #if (j<a1)
                  #declare array3[i][j]=array1[i][j];
               #else
                  #declare array3[i][j]=array2[i][j-a1];
               #end   
         #local j=j+1;
         #end
    #local i=i+1;
    #end     
#else #debug "***********Dimension mismatch (Matrix Append)***********\n"
#end
#end

// Row append

// Moore Penrose Generalised Inverse (Zhou, Zhu, Li & You Recursive LS algorithm); array2=(array1)^(-1)                                      
#macro MPGinverse(array1,array2)
 #local matm= dimension_size(array1,1); //rows
 #local matn= dimension_size(array1,2); //columns
 #local xt=array[matm][1];
 #local xttr=array[1][matn];
 #local rr=1;
 #local Pnew=array[matn][matn];
 #local Pold=array[matn][matn];
 matrixfill(Pold,matn,matn,0)
 #local Qnew=array[matn][matn];
 #local Qold=array[matn][matn];
 #local Omat=array[1][matn];
 matrixfill(Omat,1,matn,0)
 #local Knew=array[matn][1];
 #local Ktransp=array[1][matn]; 
 #local Kold=array[matn][1]; 
 #local Aux=array[matn][matn];
 #local Auxtransp=array[matn][matn]; 
 #local Auxappend=array[matn][matn+1]; 
 #local temp3=array[matn][matn];
 #local temp4=array[matn][matn];  
 #local Imat=array[matn][matn];
 identmat(Imat,matn)
 #local Qold=Imat;
 #declare Xnew=array[matn][matm];  //
 #local Xold=array[matn][matm];
 #local Xoldborder=array[matn][matm]; 
 #local test=array[1][matm];
 #local quad=0;  
 #ifndef (temp1)
 	#local temp1=array[matn][matn];
 #end	
 #ifndef (temp2)
 	#local temp2=array[matn][matn];
 #end	
 #while (rr<=matm)
       #local Xoldborder=array[rr+1][matm+1];        
       rowextr(array1,xttr,rr)
       transp(xttr,xt)
       matprod(xttr,Qold,test)
       matrixcond(test,Omat)
       #if (matrixcond1=1)
          quadform(xt,Pold,quad)
          matprod(Pold, xt, Kold)
          scalprod(Kold,Knew,1/(1+quad))
          #local Qnew=Qold;
          matprod(Knew,xttr,temp3)
          scalprod(temp3,temp4,-1)
          summat(Imat, temp4,Aux)
          matprod(Aux,Pold,Pnew)
       #else                    
          quadform(xt,Qold,quad)
          matprod(Qold, xt, Kold)
          scalprod(Kold,Knew,1/(quad))          
          matprod(Knew,xttr,temp3)
          scalprod(temp3,temp4,-1)
          summat(Imat, temp4,Aux)
          matprod(Aux,Qold,Qnew)
          transp(Aux,Auxtransp)
          transp(Knew,Ktransp)
          matprod(Knew,Ktransp,temp4)
          triplematprod(Aux,Pold,Auxtransp,temp3)
          summat(temp3, temp4,Pnew) 
       #end             
       #if (rr=1)
           #declare Xnew=Knew;
       #else    
           zzlyborder(Xold,Xoldborder)
           colappend(Aux,Knew,Auxappend)
           matprod(Auxappend,Xoldborder,Xnew)
       #end 
       
  #declare Xold=Xnew;
  #declare Pold=Pnew;
  #declare Qold=Qnew;     
  #local rr=rr+1;
  #end 
#declare array2=Xnew;
#end                      
///////////////////////////////////////////////////////////////////
    